# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Utility functions and classes for higher level RawDataWrapper access'''
import itertools
from neurom.core.dataformat import COLS
from neurom.core.dataformat import POINT_TYPE
from neurom.core.dataformat import ROOT_ID
from neurom.core.tree import Tree
from neurom.core.neuron import Neuron
from neurom.exceptions import SomaError
from neurom.exceptions import NonConsecutiveIDsError
from neurom.io.readers import load_data
from neurom.io.check import has_sequential_ids
import os


def get_soma_ids(rdw):
    '''Returns a list of IDs of points that are somas'''
    return rdw.get_ids(lambda r: r[COLS.TYPE] == POINT_TYPE.SOMA)


def get_initial_segment_ids(rdw):
    '''Returns a list of IDs of initial tree segments

    These are defined as non-soma points whose perent is a soma point.
    '''
    l = list(itertools.chain(*[rdw.get_children(s) for s in get_soma_ids(rdw)]))
    return [i for i in l if rdw.get_row(i)[COLS.TYPE] != POINT_TYPE.SOMA]


def make_tree(rdw, root_id=ROOT_ID, post_action=None):
    '''Return a tree obtained from a raw data block

    The tree contains rows of raw data.
    Args:
        rdw: a RawDataWrapper object.
        root_id: ID of the root of the tree to be built.
        post_action: optional function to run on the built tree.
    '''
    def add_children(t):
        '''Add children to a tree'''
        for c in rdw.get_children(t.value[COLS.ID]):
            child = Tree(rdw.get_row(c))
            t.add_child(child)
            add_children(child)
        return t

    head_node = Tree(rdw.get_row(root_id))
    add_children(head_node)

    if post_action is not None:
        post_action(head_node)

    return head_node


def make_neuron(raw_data, tree_action=None):
    '''Build a neuron from a raw data block

    The tree contains rows of raw data.
    Args:
        raw_data: a RawDataWrapper object.
        tree_action: optional function to run on the built trees.
    Raises: SomaError if no soma points in raw_data.
    '''
    _trees = [make_tree(raw_data, iseg, tree_action)
              for iseg in get_initial_segment_ids(raw_data)]
    _soma_pts = [raw_data.get_row(s_id) for s_id in get_soma_ids(raw_data)]
    if not _soma_pts:
        raise SomaError('No soma points found in data')
    return Neuron(_soma_pts, _trees)


def load_neuron(filename, tree_action=None):
    """
    Loads a neuron keeping a record of the filename.
    Args:
        filename: the path of the file storing morphology data
        tree_action: optional function to run on each of the neuron's
        neurite trees.
    Raises: SomaError if no soma points in data.
    Raises: NonConsecutiveIDsError if filename contains non-consecutive
    point IDs
    """

    data = load_data(filename)
    if not has_sequential_ids(data)[0]:
        raise NonConsecutiveIDsError('Non consecutive IDs found in raw data')

    nrn = make_neuron(data, tree_action)
    nrn.id = os.path.splitext(filename)[0]

    return nrn


def get_morph_files(directory):
    '''Get a list of all morphology files in a directory

    Returns:
        list with all files with extensions '.swc' or '.h5' (case insensitive)
    '''
    lsdir = [os.path.join(directory, m) for m in os.listdir(directory)]
    return [m for m in lsdir
            if os.path.isfile(m) and
            os.path.splitext(m)[1].lower() in ('.swc', '.h5')]



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

''' Quick and easy neuron morphology analysis tools

Examples:

    Load a neuron

    >>> from neurom import ezy
    >>> nrn = ezy.Neuron('some/data/path/morph_file.swc')

    Obtain some morphometrics

    >>> apical_seg_lengths = nrn.get_segment_lengths(ezy.TreeType.apical_dendrite)
    >>> axon_sec_lengths = nrn.get_section_lengths(ezy.TreeType.axon)

    View it in 2D and 3D

    >>> fig2d, ax2d = ezy.view(nrn)
    >>> fig2d.show()
    >>> fig3d, ax3d = ezy.view3d(nrn)
    >>> fig3d.show()

    Load neurons from a directory. This loads all SWC or HDF5 files it finds\
    and returns a list of neurons

    >>> import numpy as np  # For mean value calculation
    >>> nrns = ezy.load_neurons('some/data/directory')
    >>> for nrn in nrns:
    ...     print 'mean section length', np.mean([n for n in nrn.get_section_lengths()])

'''

from .neuron import Neuron
from .neuron import TreeType
from ..core.types import NEURITES as NEURITE_TYPES
from ..core.neuron import bounding_box
from ..view.view import neuron as view
from ..view.view import neuron3d as view3d
from ..io.utils import get_morph_files


def load_neurons(directory):
    '''Create a list of Neuron objects from each morphology file in directory'''
    return [Neuron(m) for m in get_morph_files(directory)]



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

''' Module for morphology HDF5 data loading

Data is unpacked into a 2-dimensional raw data block:

    [X, Y, Z, R, TYPE, ID, PARENT_ID]


HDF5.V1 Input row format:
            points: [X, Y, Z, D] (ID is position)
            groups: [FIRST_POINT_ID, TYPE, PARENT_GROUP_ID]

There is one such row per measured point.

'''
from bisect import bisect_right

import h5py
import numpy as np


def get_version(h5file):
    '''Determine whether an HDF5 file is v1 or v2

    Return: 'H5V1', 'H5V2' or None
    '''
    if 'points' in h5file and 'structure' in h5file:
        return 'H5V1'
    elif 'neuron1/structure' in h5file:
        return 'H5V2'


class H5V1(object):
    '''Read HDF5 v1 files and unpack into internal raw data block

    Input row format:
        points: (PX, PY, PZ, PD) -> [X, Y, Z, D] (ID is position)
        groups: (GPFIRST, GTYPE, GPID) -> [FIRST_POINT_ID, TYPE, PARENT_GROUP_ID]

    Internal row format: [X, Y, Z, R, TYPE, ID, PARENT_ID]
    '''

    (PX, PY, PZ, PD) = xrange(4)  # points
    (GPFIRST, GTYPE, GPID) = xrange(3)  # groups or structure


class H5(object):
    '''Read HDF5 v1 or v2 files and unpack into internal raw data block

    Internal row format: [X, Y, Z, R, TYPE, ID, PARENT_ID]
    '''

    @staticmethod
    def read_v1(filename):
        '''Read an HDF5 v1 file and return a tuple of data, offset, format.'''
        points, groups = _unpack_v1(h5py.File(filename))
        data = H5.unpack_data(points, groups)
        offset = 0  # H5 is index based, so there's no offset
        return data, offset, 'H5V1'

    @staticmethod
    def read_v2(filename, stage='raw'):
        '''Read an HDF5 v2 file and return a tuple of data, offset, format.'''
        points, groups = _unpack_v2(h5py.File(filename), stage)
        data = H5.unpack_data(points, groups)
        offset = 0  # H5 is index based, so there's no offset
        return data, offset, 'H5V2'

    @staticmethod
    def read(filename):
        '''Read a file and return a tuple of data, offset, format.

        * Tries to guess the format and the H5 version.
        * Unpacks the first block it finds out of ('repaired', 'unraveled', 'raw')
        '''
        h5file = h5py.File(filename)
        version = get_version(h5file)
        if version == 'H5V1':
            points, groups = _unpack_v1(h5py.File(filename))
        elif version == 'H5V2':
            stg = next(s for s in ('repaired', 'unraveled', 'raw')
                       if s in h5file['neuron1'])
            points, groups = _unpack_v2(h5py.File(filename), stage=stg)
        data = H5.unpack_data(points, groups)
        offset = 0  # H5 is index based, so there's no offset
        return data, offset, version

    @staticmethod
    def unpack_data(points, groups):
        '''Unpack data from h5 data groups into internal format'''

        def find_group(point_id):
            '''Find the structure group a points id belongs to

            Return: group or section point_id belongs to. Last group if
                    point_id out of bounds.
            '''
            bs = bisect_right(groups[:, H5V1.GPFIRST], point_id)
            bs = max(bs - 1, 0)
            return groups[bs]

        def find_parent_id(point_id):
            '''Find the parent ID of a point'''
            group = find_group(point_id)
            if point_id != group[H5V1.GPFIRST]:
                # point is not first point in section
                # so parent is previous point
                return point_id - 1
            else:
                # parent is last point in parent group
                parent_group_id = group[H5V1.GPID]
                # get last point in parent group
                return groups[parent_group_id + 1][H5V1.GPFIRST] - 1

        return np.array([(p[H5V1.PX], p[H5V1.PY], p[H5V1.PZ], p[H5V1.PD] / 2.,
                          find_group(i)[H5V1.GTYPE], i,
                          find_parent_id(i))
                         for i, p in enumerate(points)])


def _unpack_v1(h5file):
    '''Unpack groups from HDF5 v1 file'''
    points = np.array(h5file['points'])
    groups = np.array(h5file['structure'])
    return points, groups


def _unpack_v2(h5file, stage):
    '''Unpack groups from HDF5 v2 file'''
    points = np.array(h5file['neuron1/%s/points' % stage])
    groups = np.array(h5file['neuron1/structure/%s' % stage])
    stypes = np.array(h5file['neuron1/structure/sectiontype'])
    groups = np.hstack([groups, stypes])
    groups[:, [1, 2]] = groups[:, [2, 1]]
    return points, groups



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Python module of NeuroM to visualize morphologies
'''


from neurom.view import common
from neurom.core.types import TreeType
from matplotlib.collections import LineCollection
import numpy as np
from neurom.core.tree import isegment
from neurom.core.tree import val_iter
from neurom.io.readers import COLS
from neurom.analysis.morphtree import get_bounding_box
from neurom.analysis.morphtree import get_tree_type
from neurom.analysis.morphtree import i_segment_radius


def get_default(variable, **kwargs):
    """
    Returns default variable or kwargs variable if it exists.
    """
    default = {'linewidth': 1.2,
               'alpha': 0.8,
               'treecolor': None,
               'diameter': True,
               'diameter_scale': 1.0,
               'white_space': 30.}

    return kwargs.get(variable, default[variable])


def tree(tr, plane='xy', new_fig=True, subplot=False, **kwargs):

    """
    Generates a 2d figure of the tree.

    Parameters
    ----------
    tr: Tree
        neurom.Tree object

    Options
    -------
    plane: str
        Accepted values: Any pair of of xyz
        Default value is 'xy'.treecolor

    linewidth: float
        Defines the linewidth of the tree,
        if diameter is set to False.
        Default value is 1.2.

    alpha: float
        Defines the transparency of the tree.
        0.0 transparent through 1.0 opaque.
        Default value is 0.8.

    treecolor: str or None
        Defines the color of the tree.
        If None the default values will be used,
        depending on the type of tree:
        Basal dendrite: "red"
        Axon : "blue"
        Apical dendrite: "purple"
        Undefined tree: "black"
        Default value is None.

    new_fig: boolean
        Defines if the tree will be plotted
        in the current figure (False)
        or in a new figure (True)
        Default value is True.

    subplot: matplotlib subplot value or False
        If False the default subplot 111 will be used.
        For any other value a matplotlib subplot
        will be generated.
        Default value is False.

    diameter: boolean
        If True the diameter, scaled with diameter_scale factor,
        will define the width of the tree lines.
        If False use linewidth to select the width of the tree lines.
        Default value is True.

    diameter_scale: float
        Defines the scale factor that will be multiplied
        with the diameter to define the width of the tree line.
        Default value is 1.

    limits: list or boolean
        List of type: [[xmin, ymin, zmin], [xmax, ymax, zmax]]
        If False the figure will not be scaled.
        If True the figure will be scaled according to tree limits.
        Default value is False.

    white_space: float
        Defines the white space around
        the boundary box of the morphology.
        Default value is 1.

    Returns
    --------
    A 2D matplotlib figure with a tree view, at the selected plane.

    """
    if plane not in ('xy', 'yx', 'xz', 'zx', 'yz', 'zy'):
        return None, 'No such plane found! Please select one of: xy, xz, yx, yz, zx, zy.'

    # Initialization of matplotlib figure and axes.
    fig, ax = common.get_figure(new_fig=new_fig, subplot=subplot)

    # Data needed for the viewer: x,y,z,r
    bounding_box = get_bounding_box(tr)

    def _seg_2d(seg):
        '''2d coordinates needed for the plotting of a segment'''
        horz = getattr(COLS, plane[0].capitalize())
        vert = getattr(COLS, plane[1].capitalize())
        parent_point = seg[0]
        child_point = seg[1]
        horz1 = parent_point[horz]
        horz2 = child_point[horz]
        vert1 = parent_point[vert]
        vert2 = child_point[vert]
        return ((horz1, vert1), (horz2, vert2))

    segs = [_seg_2d(seg) for seg in val_iter(isegment(tr))]

    linewidth = get_default('linewidth', **kwargs)
    # Definition of the linewidth according to diameter, if diameter is True.
    if get_default('diameter', **kwargs):
        scale = get_default('diameter_scale', **kwargs)
        # TODO: This was originally a numpy array. Did it have to be one?
        linewidth = [2 * d * scale for d in i_segment_radius(tr)]

    # Plot the collection of lines.
    collection = LineCollection(segs,
                                color=common.get_color(get_default('treecolor', **kwargs),
                                                       get_tree_type(tr)),
                                linewidth=linewidth, alpha=get_default('alpha', **kwargs))

    ax.add_collection(collection)

    kwargs['title'] = kwargs.get('title', 'Tree view')
    kwargs['xlabel'] = kwargs.get('xlabel', plane[0])
    kwargs['ylabel'] = kwargs.get('ylabel', plane[1])
    kwargs['xlim'] = kwargs.get('xlim', [bounding_box[0][getattr(COLS, plane[0].capitalize())] -
                                         get_default('white_space', **kwargs),
                                         bounding_box[1][getattr(COLS, plane[0].capitalize())] +
                                         get_default('white_space', **kwargs)])
    kwargs['ylim'] = kwargs.get('ylim', [bounding_box[0][getattr(COLS, plane[1].capitalize())] -
                                         get_default('white_space', **kwargs),
                                         bounding_box[1][getattr(COLS, plane[1].capitalize())] +
                                         get_default('white_space', **kwargs)])

    return common.plot_style(fig=fig, ax=ax, **kwargs)


def soma(sm, plane='xy', new_fig=True, subplot=False, **kwargs):

    """
    Generates a 2d figure of the soma.

    Parameters
    ----------
    soma: Soma
        neurom.Soma object

    Options
    -------
    plane: str
        Accepted values: Any pair of of xyz
        Default value is 'xy'

    linewidth: float
        Defines the linewidth of the soma.
        Default value is 1.2

    alpha: float
        Defines the transparency of the soma.
        0.0 transparent through 1.0 opaque.
        Default value is 0.8.

    treecolor: str or None
        Defines the color of the soma.
        If None the default value will be used:
        Soma : "black".
        Default value is None.

    new_fig: boolean
        Defines if the tree will be plotted
        in the current figure (False)
        or in a new figure (True)
        Default value is True.

    subplot: matplotlib subplot value or False
        If False the default subplot 111 will be used.
        For any other value a matplotlib subplot
        will be generated.
        Default value is False.

    limits: list or boolean
        List of type: [[xmin, ymin, zmin], [xmax, ymax, zmax]]
        If False the figure will not be scaled.
        If True the figure will be scaled according to tree limits.
        Default value is False.

    Returns
    --------
    A 2D matplotlib figure with a soma view, at the selected plane.

    """

    treecolor = kwargs.get('treecolor', None)
    outline = kwargs.get('outline', True)

    if plane not in ('xy', 'yx', 'xz', 'zx', 'yz', 'zy'):
        return None, 'No such plane found! Please select one of: xy, xz, yx, yz, zx, zy.'

    # Initialization of matplotlib figure and axes.
    fig, ax = common.get_figure(new_fig=new_fig, subplot=subplot)

    # Definition of the tree color depending on the tree type.
    treecolor = common.get_color(treecolor, tree_type=TreeType.soma)

    # Plot the outline of the soma as a circle, is outline is selected.
    if not outline:
        soma_circle = common.plt.Circle(sm.center, sm.radius, color=treecolor,
                                        alpha=get_default('alpha', **kwargs))
        ax.add_artist(soma_circle)
    else:
        horz = []
        vert = []

        for s_point in sm.iter():
            horz.append(s_point[getattr(COLS, plane[0].capitalize())])
            vert.append(s_point[getattr(COLS, plane[1].capitalize())])

        horz.append(horz[0]) # To close the loop for a soma viewer. This might be modified!
        vert.append(vert[0]) # To close the loop for a soma viewer. This might be modified!

        common.plt.plot(horz, vert, color=treecolor,
                        alpha=get_default('alpha', **kwargs),
                        linewidth=get_default('linewidth', **kwargs))

    kwargs['title'] = kwargs.get('title', 'Soma view')
    kwargs['xlabel'] = kwargs.get('xlabel', plane[0])
    kwargs['ylabel'] = kwargs.get('ylabel', plane[1])

    return common.plot_style(fig=fig, ax=ax, **kwargs)


def neuron(nrn, plane='xy', new_fig=True, subplot=False, **kwargs):

    """
    Generates a 2d figure of the neuron,
    that contains a soma and a list of trees.

    Parameters
    ----------
    neuron: Neuron
        neurom.Neuron object

    Options
    -------
    plane: str
        Accepted values: Any pair of of xyz
        Default value is 'xy'

    linewidth: float
        Defines the linewidth of the tree and soma
        of the neuron, if diameter is set to False.
        Default value is 1.2.

    alpha: float
        Defines the transparency of the neuron.
        0.0 transparent through 1.0 opaque.
        Default value is 0.8.

    treecolor: str or None
        Defines the color of the trees.
        If None the default values will be used,
        depending on the type of tree:
        Soma: "black"
        Basal dendrite: "red"
        Axon : "blue"
        Apical dendrite: "purple"
        Undefined tree: "black"
        Default value is None.

    new_fig: boolean
        Defines if the neuron will be plotted
        in the current figure (False)
        or in a new figure (True)
        Default value is True.

    subplot: matplotlib subplot value or False
        If False the default subplot 111 will be used.
        For any other value a matplotlib subplot
        will be generated.
        Default value is False.

    diameter: boolean
        If True the diameter, scaled with diameter_scale factor,
        will define the width of the tree lines.
        If False use linewidth to select the width of the tree lines.
        Default value is True.

    diameter_scale: float
        Defines the scale factor that will be multiplied
        with the diameter to define the width of the tree line.
        Default value is 1.

    limits: list or boolean
        List of type: [[xmin, ymin, zmin], [xmax, ymax, zmax]]
        If False the figure will not be scaled.
        If True the figure will be scaled according to tree limits.
        Default value is False.

    Returns
    --------
    A 3D matplotlib figure with a tree view, at the selected plane.

    """
    if plane not in ('xy', 'yx', 'xz', 'zx', 'yz', 'zy'):
        return None, 'No such plane found! Please select one of: xy, xz, yx, yz, zx, zy.'

    new_fig = kwargs.get('new_fig', True)
    subplot = kwargs.get('subplot', False)

    # Initialization of matplotlib figure and axes.
    fig, ax = common.get_figure(new_fig=new_fig, subplot=subplot)

    kwargs['new_fig'] = False

    soma(nrn.soma, plane=plane, **kwargs)

    h = []
    v = []

    for temp_tree in nrn.neurites:

        bounding_box = get_bounding_box(temp_tree)

        h.append([bounding_box[0][getattr(COLS, plane[0].capitalize())],
                  bounding_box[1][getattr(COLS, plane[0].capitalize())]])
        v.append([bounding_box[0][getattr(COLS, plane[1].capitalize())],
                  bounding_box[1][getattr(COLS, plane[1].capitalize())]])

        tree(temp_tree, plane=plane, **kwargs)

    kwargs['title'] = kwargs.get('title', 'Neuron view')
    kwargs['xlabel'] = kwargs.get('xlabel', plane[0])
    kwargs['ylabel'] = kwargs.get('ylabel', plane[1])
    kwargs['xlim'] = kwargs.get('xlim', [np.min(h) - get_default('white_space', **kwargs),
                                         np.max(h) + get_default('white_space', **kwargs)])
    kwargs['ylim'] = kwargs.get('ylim', [np.min(v) - get_default('white_space', **kwargs),
                                         np.max(v) + get_default('white_space', **kwargs)])

    return common.plot_style(fig=fig, ax=ax, **kwargs)


def tree3d(tr, new_fig=True, new_axes=True, subplot=False, **kwargs):

    """
    Generates a figure of the tree in 3d.

    Parameters
    ----------
    tr: Tree
        neurom.Tree object

    Options
    -------
    linewidth: float
        Defines the linewidth of the tree,
        if diameter is set to False.
        Default value is 1.2.

    alpha: float
        Defines the transparency of the tree.
        0.0 transparent through 1.0 opaque.
        Default value is 0.8.

    treecolor: str or None
        Defines the color of the tree.
        If None the default values will be used,
        depending on the type of tree:
        Basal dendrite: "red"
        Axon : "blue"
        Apical dendrite: "purple"
        Undefined tree: "black"
        Default value is None.

    new_fig: boolean
        Defines if the tree will be plotted
        in the current figure (False)
        or in a new figure (True)
        Default value is True.

    subplot: matplotlib subplot value or False
        If False the default subplot 111 will be used.
        For any other value a matplotlib subplot
        will be generated.
        Default value is False.

    diameter: boolean
        If True the diameter, scaled with diameter_scale factor,
        will define the width of the tree lines.
        If False use linewidth to select the width of the tree lines.
        Default value is True.

    diameter_scale: float
        Defines the scale factor that will be multiplied
        with the diameter to define the width of the tree line.
        Default value is 1.

    white_space: float
        Defines the white space around
        the boundary box of the morphology.
        Default value is 1.

    Returns
    --------
    A 3D matplotlib figure with a tree view.

    """
    from mpl_toolkits.mplot3d.art3d import Line3DCollection

    # Initialization of matplotlib figure and axes.
    fig, ax = common.get_figure(new_fig=new_fig, new_axes=new_axes,
                                subplot=subplot, params={'projection': '3d'})

    # Data needed for the viewer: x,y,z,r
    bounding_box = get_bounding_box(tr)

    def _seg_3d(seg):
        '''2d coordinates needed for the plotting of a segment'''
        horz = getattr(COLS, 'X')
        vert = getattr(COLS, 'Y')
        depth = getattr(COLS, 'Z')
        parent_point = seg[0]
        child_point = seg[1]
        horz1 = parent_point[horz]
        horz2 = child_point[horz]
        vert1 = parent_point[vert]
        vert2 = child_point[vert]
        depth1 = parent_point[depth]
        depth2 = child_point[depth]
        return ((horz1, vert1, depth1), (horz2, vert2, depth2))

    segs = [_seg_3d(seg) for seg in val_iter(isegment(tr))]

    linewidth = get_default('linewidth', **kwargs)

    # Definition of the linewidth according to diameter, if diameter is True.
    if get_default('diameter', **kwargs):
        # TODO: This was originally a numpy array. Did it have to be one?
        linewidth = [2 * d * get_default('diameter_scale', **kwargs) for d in i_segment_radius(tr)]

    # Plot the collection of lines.
    collection = Line3DCollection(segs,
                                  color=common.get_color(get_default('treecolor', **kwargs),
                                                         get_tree_type(tr)),
                                  linewidth=linewidth, alpha=get_default('alpha', **kwargs))

    ax.add_collection3d(collection)

    kwargs['title'] = kwargs.get('title', 'Tree 3d-view')
    kwargs['xlabel'] = kwargs.get('xlabel', 'X')
    kwargs['ylabel'] = kwargs.get('ylabel', 'Y')
    kwargs['zlabel'] = kwargs.get('zlabel', 'Z')
    kwargs['xlim'] = kwargs.get('xlim', [bounding_box[0][0] - get_default('white_space', **kwargs),
                                         bounding_box[1][0] + get_default('white_space', **kwargs)])
    kwargs['ylim'] = kwargs.get('ylim', [bounding_box[0][1] - get_default('white_space', **kwargs),
                                         bounding_box[1][1] + get_default('white_space', **kwargs)])
    kwargs['zlim'] = kwargs.get('zlim', [bounding_box[0][2] - get_default('white_space', **kwargs),
                                         bounding_box[1][2] + get_default('white_space', **kwargs)])

    return common.plot_style(fig=fig, ax=ax, **kwargs)


def soma3d(sm, new_fig=True, new_axes=True, subplot=False, **kwargs):

    """
    Generates a 3d figure of the soma.

    Parameters
    ----------
    soma: Soma
        neurom.Soma object

    Options
    -------
    alpha: float
        Defines the transparency of the soma.
        0.0 transparent through 1.0 opaque.
        Default value is 0.8.

    treecolor: str or None
        Defines the color of the soma.
        If None the default value will be used:
        Soma : "black".
        Default value is None.

    new_fig: boolean
        Defines if the tree will be plotted
        in the current figure (False)
        or in a new figure (True)
        Default value is True.

    subplot: matplotlib subplot value or False
        If False the default subplot 111 will be used.
        For any other value a matplotlib subplot
        will be generated.
        Default value is False.

    Returns
    --------
    A 3D matplotlib figure with a soma view.

    """

    treecolor = kwargs.get('treecolor', None)

    # Initialization of matplotlib figure and axes.
    fig, ax = common.get_figure(new_fig=new_fig, new_axes=new_axes,
                                subplot=subplot, params={'projection': '3d'})

    # Definition of the tree color depending on the tree type.
    treecolor = common.get_color(treecolor, tree_type=TreeType.soma)

    xs = sm.center[0]
    ys = sm.center[1]
    zs = sm.center[2]

    # Plot the soma as a circle.
    fig, ax = common.plot_sphere(fig, ax, center=[xs, ys, zs], radius=sm.radius, color=treecolor,
                                 alpha=get_default('alpha', **kwargs))

    kwargs['title'] = kwargs.get('title', 'Soma view')
    kwargs['xlabel'] = kwargs.get('xlabel', 'X')
    kwargs['ylabel'] = kwargs.get('ylabel', 'Y')
    kwargs['zlabel'] = kwargs.get('zlabel', 'Z')

    return common.plot_style(fig=fig, ax=ax, **kwargs)


def neuron3d(nrn, new_fig=True, new_axes=True, subplot=False, **kwargs):

    """
    Generates a figure of the neuron,
    that contains a soma and a list of trees.

    Parameters
    ----------
    neuron: Neuron
        neurom.Neuron object

    Options
    -------
    linewidth: float
        Defines the linewidth of the tree and soma
        of the neuron, if diameter is set to False.
        Default value is 1.2.

    alpha: float
        Defines the transparency of the neuron.
        0.0 transparent through 1.0 opaque.
        Default value is 0.8.

    treecolor: str or None
        Defines the color of the trees.
        If None the default values will be used,
        depending on the type of tree:
        Soma: "black"
        Basal dendrite: "red"
        Axon : "blue"
        Apical dendrite: "purple"
        Undefined tree: "black"
        Default value is None.

    new_fig: boolean
        Defines if the neuron will be plotted
        in the current figure (False)
        or in a new figure (True)
        Default value is True.

    subplot: matplotlib subplot value or False
        If False the default subplot 111 will be used.
        For any other value a matplotlib subplot
        will be generated.
        Default value is False.

    diameter: boolean
        If True the diameter, scaled with diameter_scale factor,
        will define the width of the tree lines.
        If False use linewidth to select the width of the tree lines.
        Default value is True.

    diameter_scale: float
        Defines the scale factor that will be multiplied
        with the diameter to define the width of the tree line.
        Default value is 1.

    Returns
    --------
    A 3D matplotlib figure with a neuron view.

    """
    new_fig = kwargs.get('new_fig', True)
    subplot = kwargs.get('subplot', False)

    # Initialization of matplotlib figure and axes.
    fig, ax = common.get_figure(new_fig=new_fig, new_axes=new_axes,
                                subplot=subplot, params={'projection': '3d'})

    kwargs['new_fig'] = False
    kwargs['new_axes'] = False

    soma3d(nrn.soma, **kwargs)

    h = []
    v = []
    d = []

    for temp_tree in nrn.neurites:

        bounding_box = get_bounding_box(temp_tree)

        h.append([bounding_box[0][getattr(COLS, 'X')],
                  bounding_box[1][getattr(COLS, 'X')]])
        v.append([bounding_box[0][getattr(COLS, 'Y')],
                  bounding_box[1][getattr(COLS, 'Y')]])
        d.append([bounding_box[0][getattr(COLS, 'Z')],
                  bounding_box[1][getattr(COLS, 'Z')]])

        tree3d(temp_tree, **kwargs)

    kwargs['title'] = kwargs.get('title', 'Neuron view')

    kwargs['xlim'] = kwargs.get('xlim', [np.min(h) - get_default('white_space', **kwargs),
                                         np.max(h) + get_default('white_space', **kwargs)])
    kwargs['ylim'] = kwargs.get('ylim', [np.min(v) - get_default('white_space', **kwargs),
                                         np.max(v) + get_default('white_space', **kwargs)])
    kwargs['zlim'] = kwargs.get('zlim', [np.min(d) - get_default('white_space', **kwargs),
                                         np.max(d) + get_default('white_space', **kwargs)])

    return common.plot_style(fig=fig, ax=ax, **kwargs)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Module containing the common functionality
to be used by view-plot modules.
"""
from neurom.core.types import TreeType
import os
import matplotlib

#  Awful hack to use non-GUI backend when no display
#  is available. For unixy systems.
if 'DISPLAY' not in os.environ: # noqa
    matplotlib.use('Agg')

import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.mplot3d import Axes3D  # pylint: disable=unused-import


# Map tree type to color
TREE_COLOR = {TreeType.basal_dendrite: 'red',
              TreeType.apical_dendrite: 'purple',
              TreeType.axon: 'blue',
              TreeType.soma: 'black',
              TreeType.undefined: 'green'}


def figure_naming(pretitle=None, posttitle=None, prefile=None, postfile=None):
    """
    Helping function to define the strings that handle pre-post conventions
    for viewing - plotting title and saving options.

    Options
    ----------
    pretitle : str
        String to include before the general title of the figure.
        Default is None.

    posttitle : str
        String to include after the general title of the figure.
        Default is None.

    prefile : str
        String to include before the general filename of the figure.
        Default is None.

    postfile : str
        String to include after the general filename of the figure.
        Default is None.

    Returns
    -------
    string_output : string
        String to include in the figure name and title, in a suitable form.

    """
    if not pretitle:
        pretitle = ""
    else:
        pretitle = "%s -- " % pretitle

    if not posttitle:
        posttitle = ""
    else:
        posttitle = " -- %s" % posttitle

    if not prefile:
        prefile = ""
    else:
        prefile = "%s_" % prefile

    if not postfile:
        postfile = ""
    else:
        postfile = "_%s" % postfile

    return pretitle, posttitle, prefile, postfile


def get_color(treecolor, tree_type):
    """
    If treecolor is None returns the color depending on the type:
    Basal dendrite: "red"
    Axon : "blue"
    Apical dendrite: "purple"
    Soma tree: "black"
    Undefined or any other type: "green"
    Otherwise returns the treecolor.
    """
    if treecolor is not None:
        return treecolor
    else:
        return TREE_COLOR.get(tree_type, 'green')


def get_figure(new_fig=True, new_axes=True, subplot=False, params=None, no_axes=False):
    """
    Function to be used for viewing - plotting,
    to initialize the matplotlib figure - axes.

    Options
    ----------
    new_fig : boolean
        Defines if a new figure will be created.
        If False the current figure is returned.
        Default is True.

    subplot : subplot or False
        Defines if a subplot will be generated.
        If False the subplot is the standard 111.
        Default is False.

    params: dict
        If empty dictonary no supplementary parameters will be used.
        If dict not empty the parameters will be passed to the initialization of the axes.
        Default value is {}

    no_axes: boolean
        Defines the output of the function:
        If False the axes is returned.
        If True the figure is returned.
        Default value is False.

    Returns
    -------
    figure or axes
        Returns a figure if no_axes is True,
        or axes if no_axes is False.

    """
    if new_fig:
        fig = plt.figure()
    else:
        fig = plt.gcf()

    if no_axes:
        return fig

    if not subplot:
        subplot = 111

    if params is None:
        params = dict()

    if new_axes:
        if isinstance(subplot, (tuple, list)):
            ax = fig.add_subplot(subplot[0], subplot[1], subplot[2], **params)
        else:
            ax = fig.add_subplot(subplot, **params)
    else:
        ax = plt.gca()

    return fig, ax


def save_plot(fig, **kwargs):

    """
    Function to be used for viewing - plotting
    to save a matplotlib figure.

    Input
    -------
    fig: matplotlib figure

    Options
    ----------
    prefile : str
        String to include before the general filename of the figure.
        Default is None.

    postfile : str
        String to include after the general filename of the figure.
        Default is None.

    output_path : str
        String to define the path to the output directory.
        Default value is './'

    output_name : str
        String to define the name of the output figure.
        Default value is 'Figure'

    output_format : str
        String to define the format of the output figure.
        Default value is 'png'

    dpi: int
        Define the DPI (Dots per Inch) of the figure.
        Default value is 300.

    transparent: boolean
        If True the saved figure will have a transparent background.
        Default value is False.

    Returns
    -------
    Generates a figure file in the selected directory.

    """

    prefile = kwargs.get('prefile', '')
    postfile = kwargs.get('postfile', '')
    output_path = kwargs.get('output_path', './')
    output_name = kwargs.get('output_name', 'Figure')
    output_format = kwargs.get('output_format', 'png')
    dpi = kwargs.get('dpi', 300)
    transparent = kwargs.get('transparent', False)

    if not os.path.exists(output_path):
        os.makedirs(output_path) # Make output directory if non-exsiting

    output = os.path.join(output_path, prefile + output_name + postfile + "." + output_format)

    plt.savefig(output, dpi=dpi, transparent=transparent)

    print 'Plot Saved:', output

    return fig


def plot_style(fig, ax, **kwargs):

    """
    Function to set the basic options of a matplotlib figure,
    to be used by viewing - plotting functions.

    Parameters
    ----------
    ax: matplotlib axes

    Options
    -------
    pretitle : str
        String to include before the general title of the figure.
        Default value is None.

    posttitle : str
        String to include after the general title of the figure.
        Default value is None.

    title : str
        The title for the figure
        Default value is "Figure".

    no_title : boolean
        Defines the presence of a title in the figure.
        If True the title will be set to an empty string "".
        Default value is False.

    title_fontsize : int
        Defines the size of the title's font.
        Default value is 14.

    title_arg : dict
        Defines the arguments that will be passsed
        into matplotlib as title arguments.
        Default value is None.

    xlabel : str
        The xlabel for the figure
        Default value is "X".

    ylabel : str
        The xlabel for the figure
        Default value is "Y".

    no_labels : boolean
        Defines the presence of the labels in the figure.
        If True the labels will be set to an empty string "".
        Default value is False.

    no_xlabel : boolean
        Defines the presence of the xlabel in the figure.
        If True the xlabel will be set to an empty string "".
        Default value is False.

    no_ylabel : boolean
        Defines the presence of the ylabel in the figure.
        If True the ylabel will be set to an empty string "".
        Default value is False.

    label_fontsize : int
        Defines the size of the labels' font.
        Default value is 14.

    xlabel_arg : dict
        Defines the arguments that will be passsed
        into matplotlib as xlabel arguments.
        Default value is None.

    ylabel_arg : dict
        Defines the arguments that will be passsed
        into matplotlib as ylabel arguments.
        Default value is None.

    no_ticks : boolean
        Defines the presence of x-y ticks in the figure.
        If True the ticks will be empty.
        Default value is False.

    no_xticks : boolean
        Defines the presence of x ticks in the figure.
        If True the ticks will be empty.
        Default value is False.

    no_yticks : boolean
        Defines the presence of y ticks in the figure.
        If True the ticks will be empty.
        Default value is False.

    xticks : list of ticks
        Defines the values of x ticks in the figure.
        Default value is None.

    yticks : list of ticks
        Defines the values of y ticks in the figure.
        Default value is None.

    tick_fontsize : int
        Defines the size of the ticks' font.
        Default value is 12.

    xticks_arg : dict
        Defines the arguments that will be passsed
        into matplotlib as xticks arguments.
        Default value is None.

    yticks_arg : dict
        Defines the arguments that will be passsed
        into matplotlib as yticks arguments.
        Default value is None.

    no_limits : boolean
        Defines the presence of plot limits in the figure.
        Default value is False.

    no_xlim : boolean
        Defines the presence of plot x-limits in the figure.
        Default value is False.

    no_ylim : boolean
        Defines the presence of plot y-limits in the figure.
        Default value is False.

    xlim: list of two floats
        Defines the min and the max values in x-axis.
        Default in None

    ylim: list of two floats
        Defines the min and the max values in y-axis.
        Default in None

    no_legend : boolean
        Defines the presence of a legend in the figure.
        If True the legend will not be included in the Figure.
        Default value is True.

    legend_arg : dict
        Defines the arguments that will be passsed
        into matplotlib as legend arguments.
        Default value is None.

    prefile : str
        String to include before the general filename of the figure.
        Default is None.

    postfile : str
        String to include after the general filename of the figure.
        Default is None.

    output_path : str
        String to define the path to the output directory.
        Default value is './'

    output_name : str
        String to define the name of the output figure.
        Default value is 'Figure'

    output_format : str
        String to define the format of the output figure.
        Default value is 'png'

    dpi: int
        Define the DPI (Dots per Inch) of the figure.
        Default value is 300.

    transparent: boolean
        If True the saved figure will have a transparent background.
        Default value is False.

    show_plot : boolean
        If True the figure is displayed.
        Default value is True.

    no_axes: boolean
        If True the labels and the frame will be set off.
        Default value is False.

    tight: boolean
        If True the set layout of matplotlib will be activated.
        Default value is False.

    Returns
    -------
    figure_output : figure, axes

    """
    # Definition of title/file naming variables
    prefile = kwargs.get('prefile', '')
    postfile = kwargs.get('postfile', '')
    pretitle = kwargs.get('pretitle', '')
    posttitle = kwargs.get('posttitle', '')

    # Definition of global options
    no_axes = kwargs.get('no_axes', False)
    show_plot = kwargs.get('show_plot', True)
    tight = kwargs.get('tight', False)

    # Definition of save options
    output_path = kwargs.get('output_path', None)

    pretitle, posttitle, prefile, postfile = figure_naming(pretitle, posttitle, prefile, postfile)

    fig, ax = plot_title(fig, ax, **kwargs)

    fig, ax = plot_labels(fig, ax, **kwargs)

    fig, ax = plot_ticks(fig, ax, **kwargs)

    fig, ax = plot_limits(fig, ax, **kwargs)

    fig, ax = plot_legend(fig, ax, **kwargs)

    if no_axes:
        ax.set_frame_on(False)
        ax.xaxis.set_visible(False)
        ax.yaxis.set_visible(False)

    if tight:
        fig.set_tight_layout(True)

    if output_path is not None:
        fig = save_plot(fig=ax, **kwargs)

    if not show_plot:
        plt.close()
        return (None, None)

    return fig, ax


def plot_title(fig, ax, **kwargs):

    """
    Function that defines the title options
    of a matplotlib plot.

    Parameters
    ----------
    fig: matplotlib figure

    ax: matplotlib axes

    Options
    -------
    pretitle : str
        String to include before the general title of the figure.
        Default value is None.

    posttitle : str
        String to include after the general title of the figure.
        Default value is None.

    title : str
        Set the title for the figure.
        If "" no title will be added.
        Default value is "Figure".

    title_fontsize : int
        Defines the size of the title's font.
        Default value is 14.

    title_arg : dict
        Defines the arguments that will be passsed
        into matplotlib as title arguments.
        Default value is None.

    Returns
    -------
    figure_output : figure, axes

    """

    # Definition of title options
    pretitle = kwargs.get('pretitle', '')
    posttitle = kwargs.get('posttitle', '')
    title = kwargs.get('title', 'Figure')
    title_fontsize = kwargs.get('titlefontsize', 14)
    title_arg = kwargs.get('titlearg', None)

    if title_arg is None:
        title_arg = {}

    ax.set_title(pretitle + title + posttitle,
                 fontsize=title_fontsize, **title_arg)

    return fig, ax


def plot_labels(fig, ax, **kwargs):

    """
    Function that defines the labels options
    of a matplotlib plot.

    Parameters
    ----------
    fig: matplotlib figure

    ax: matplotlib axes

    Options
    -------
    xlabel : str
        The xlabel for the figure.
        For no_xlabel set to ''.
        Default value is "X".

    ylabel : str
        The xlabel for the figure.
        For no_ylabel set to ''.
        Default value is "Y".

    zlabel : str
        The zlabel for the figure.
        For no_zlabel set to ''.
        Default value is "Z".

    label_fontsize : int
        Defines the size of the labels' font.
        Default value is 14.

    xlabel_arg : dict
        Defines the arguments that will be passsed
        into matplotlib as xlabel arguments.
        Default value is None.

    ylabel_arg : dict
        Defines the arguments that will be passsed
        into matplotlib as ylabel arguments.
        Default value is None.

    zlabel_arg : dict
        Defines the arguments that will be passsed
        into matplotlib as zlabel arguments.
        Default value is None.

    Returns
    -------
    figure_output : figure, axes

    """

    # Definition of label options
    xlabel = kwargs.get('xlabel', 'X')
    ylabel = kwargs.get('ylabel', 'Y')
    zlabel = kwargs.get('zlabel', 'Z')
    label_fontsize = kwargs.get('labelfontsize', 14)
    xlabel_arg = kwargs.get('xlabel_arg', None)
    ylabel_arg = kwargs.get('ylabel_arg', None)
    zlabel_arg = kwargs.get('zlabel_arg', None)

    if xlabel_arg is None:
        xlabel_arg = {}

    if ylabel_arg is None:
        ylabel_arg = {}

    if zlabel_arg is None:
        zlabel_arg = {}

    ax.set_xlabel(xlabel, fontsize=label_fontsize, **xlabel_arg)
    ax.set_ylabel(ylabel, fontsize=label_fontsize, **ylabel_arg)

    if hasattr(ax, 'zaxis'):
        ax.set_zlabel(zlabel, fontsize=label_fontsize, **zlabel_arg)

    return fig, ax


def plot_ticks(fig, ax, **kwargs):

    """
    Function that defines the labels options
    of a matplotlib plot.

    Parameters
    ----------
    fig: matplotlib figure

    ax: matplotlib axes

    Options
    -------
    xticks : list of ticks
        Defines the values of x ticks in the figure.
        If None the xticks will not be modified.
        For no_xticks set to [].
        Default value is None.

    yticks : list of ticks
        Defines the values of y ticks in the figure.
        If None the yticks will not be modified.
        For no_yticks set to [].
        Default value is None.

    zticks : list of ticks
        Defines the values of z ticks in the figure.
        If None the zticks will not be modified.
        For no_zticks set to [].
        Default value is None.

    tick_fontsize : int
        Defines the size of the ticks' font.
        Default value is 12.

    xticks_arg : dict
        Defines the arguments that will be passsed
        into matplotlib as xticks arguments.
        Default value is None.

    yticks_arg : dict
        Defines the arguments that will be passsed
        into matplotlib as yticks arguments.
        Default value is None.

    zticks_arg : dict
        Defines the arguments that will be passsed
        into matplotlib as zticks arguments.
        Default value is None.

    Returns
    -------
    figure_output : figure, axes

    """

    # Definition of tick options
    xticks = kwargs.get('xticks', None)
    yticks = kwargs.get('yticks', None)
    zticks = kwargs.get('zticks', None)
    tick_fontsize = kwargs.get('tickfontsize', 12)
    xticks_arg = kwargs.get('xticksarg', None)
    yticks_arg = kwargs.get('yticksarg', None)
    zticks_arg = kwargs.get('zticksarg', None)

    if xticks_arg is None:
        xticks_arg = {}

    if yticks_arg is None:
        yticks_arg = {}

    if zticks_arg is None:
        zticks_arg = {}

    if xticks is not None:
        ax.set_xticks(xticks)
        ax.xaxis.set_tick_params(labelsize=tick_fontsize, **xticks_arg)

    if yticks is not None:
        ax.set_yticks(yticks)
        ax.yaxis.set_tick_params(labelsize=tick_fontsize, **yticks_arg)

    if zticks is not None:
        ax.set_zticks(zticks)
        ax.zaxis.set_tick_params(labelsize=tick_fontsize, **zticks_arg)

    return fig, ax


def plot_limits(fig, ax, **kwargs):

    """
    Function that defines the limit options
    of a matplotlib plot.

    Parameters
    ----------
    fig: matplotlib figure

    ax: matplotlib axes

    Options
    -------
    no_limits : boolean
        Defines the presence of plot limits in the figure.
        Default value is False.

    no_xlim : boolean
        Defines the presence of plot x-limits in the figure.
        Default value is False.

    no_ylim : boolean
        Defines the presence of plot y-limits in the figure.
        Default value is False.

    xlim: list of two floats
        Defines the min and the max values in x-axis.
        Default in None

    ylim: list of two floats
        Defines the min and the max values in y-axis.
        Default in None

    Returns
    -------
    figure_output : figure, axes

    """
    # Definition of limit options
    no_xlim = kwargs.get('no_xlim', False)
    no_ylim = kwargs.get('no_ylim', False)
    no_zlim = kwargs.get('no_zlim', False)
    xlim = kwargs.get('xlim', None)
    ylim = kwargs.get('ylim', None)
    zlim = kwargs.get('zlim', None)

    if not no_xlim:
        ax.set_xlim(xlim)
    if not no_ylim:
        ax.set_ylim(ylim)
    if hasattr(ax, 'zaxis') and not no_zlim:
        ax.set_zlim(zlim)

    return fig, ax


def plot_legend(fig, ax, **kwargs):

    """
    Function that defines the legend options
    of a matplotlib plot.

    Parameters
    ----------
    fig: matplotlib figure

    ax: matplotlib axes

    Options
    -------
    no_legend : boolean
        Defines the presence of a legend in the figure.
        If True the legend will not be included in the Figure.
        Default value is True.

    legend_arg : dict
        Defines the arguments that will be passsed
        into matplotlib as legend arguments.
        Default value is None.

    Returns
    -------
    figure_output : figure, axes

    """
    # Definition of legend options
    no_legend = kwargs.get('no_legend', True)
    legend_arg = kwargs.get('legendarg', None)

    if legend_arg is None:
        legend_arg = {}

    if not no_legend:
        ax.legend(**legend_arg)

    return fig, ax


def plot_sphere(fig, ax, center, radius, color='black', alpha=1.):
    """
    Plots a 3d sphere, given the center and the radius.
    """

    u = np.linspace(0, 2 * np.pi, 300)
    v = np.linspace(0, np.pi, 300)

    x = center[0] + radius * np.outer(np.cos(u), np.sin(v))
    y = center[1] + radius * np.outer(np.sin(u), np.sin(v))
    z = center[2] + radius * np.outer(np.ones_like(u), np.cos(v))

    ax.plot_surface(x, y, z, linewidth=0.0, color=color, alpha=alpha)

    return fig, ax



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Python module of NeuroM to check neurons.
'''

from neurom.core.types import TreeType
from neurom.core.tree import ipreorder
from neurom.core.tree import isegment
from neurom.core.tree import isection
from neurom.core.tree import val_iter
from neurom.core.dataformat import COLS
from neurom.analysis.morphmath import section_length
from neurom.analysis.morphmath import segment_length
from neurom.analysis.morphtree import find_tree_type
from itertools import chain


def has_axon(neuron, treefun=find_tree_type):
    '''Check if a neuron has an axon

    Arguments:
        neuron: The neuron object to test
        treefun: Optional function to calculate the tree type of
        neuron's neurites
    '''
    return TreeType.axon in [treefun(n) for n in neuron.neurites]


def has_apical_dendrite(neuron, min_number=1, treefun=find_tree_type):
    '''Check if a neuron has apical dendrites

    Arguments:
        neuron: The neuron object to test
        min_number: minimum number of apical dendrites required
        treefun: Optional function to calculate the tree type of neuron's
        neurites
    '''
    types = [treefun(n) for n in neuron.neurites]
    return types.count(TreeType.apical_dendrite) >= min_number


def has_basal_dendrite(neuron, min_number=1, treefun=find_tree_type):
    '''Check if a neuron has basal dendrites

    Arguments:
        neuron: The neuron object to test
        min_number: minimum number of basal dendrites required
        treefun: Optional function to calculate the tree type of neuron's
        neurites
    '''
    types = [treefun(n) for n in neuron.neurites]
    return types.count(TreeType.basal_dendrite) >= min_number


def nonzero_segment_lengths(neuron, threshold=0.0):
    '''Check presence of neuron segments with length not above threshold

    Arguments:
        neuron: Neuron object whose segments will be tested
        threshold: value above which a segment length is considered to be non-zero
    Return: list of (first_id, second_id) of zero length segments
    '''
    l = [[s for s in val_iter(isegment(t))
          if segment_length(s) <= threshold]
         for t in neuron.neurites]
    return [(i[0][COLS.ID], i[1][COLS.ID]) for i in chain(*l)]


def nonzero_section_lengths(neuron, threshold=0.0):
    '''Check presence of neuron sections with length not above threshold

    Arguments:
        neuron: Neuron object whose segments will be tested
        threshold: value above which a section length is considered to be non-zero
    Return: list of ids of first point in bad sections
    '''
    l = [[s for s in val_iter(isection(t))
          if section_length(s) <= threshold]
         for t in neuron.neurites]
    return [i[0][COLS.ID] for i in chain(*l)]


def nonzero_neurite_radii(neuron, threshold=0.0):
    '''Check presence of neurite points with radius not above threshold

    Arguments:
        neuron: Neuron object whose segments will be tested
        threshold: value above which a radius is considered to be non-zero
    Return: list of IDs of zero-radius points
    '''

    ids = [[i[COLS.ID] for i in val_iter(ipreorder(t))
            if i[COLS.R] <= threshold] for t in neuron.neurites]
    return [i for i in chain(*ids)]



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Module with consistency/validity checks for raw data  blocks'''
import numpy as np
from neurom.core.dataformat import ROOT_ID
from neurom.core.dataformat import COLS
from neurom.core.dataformat import POINT_TYPE


def has_sequential_ids(raw_data):
    '''Check that IDs are increasing and consecutive

    returns tuple (bool, list of IDs that are not consecutive
    with their predecessor)
    '''
    ids = raw_data.get_col(COLS.ID)
    steps = [int(j) for (i, j) in zip(ids, ids[1:]) if int(j - i) != 1]
    return len(steps) == 0, steps


def has_soma_points(raw_data):
    '''Checks if the TYPE column of raw data block has
    an element of type soma'''
    return POINT_TYPE.SOMA in raw_data.get_col(COLS.TYPE)


def has_all_finite_radius_neurites(raw_data):
    '''Check that all points with neurite type have a finite radius

    return: tuple of (bool, [IDs of neurite points with zero radius])
    '''
    bad_pts = [int(row[COLS.ID]) for row in raw_data.iter_row()
               if row[COLS.TYPE] in POINT_TYPE.NEURITES and row[COLS.R] == 0.0]
    return len(bad_pts) == 0, bad_pts


def is_neurite_segment(segment):
    '''Check that both points in a segment are of neurite type

    argument:
        segment: pair of raw data rows representing a segment

    return: true if both have neurite type
    '''
    return (segment[0][COLS.TYPE] in POINT_TYPE.NEURITES and
            segment[1][COLS.TYPE] in POINT_TYPE.NEURITES)


def has_all_finite_length_segments(raw_data):
    '''Check that all segments of neurite type have a finite length

    return: tuple of (bool, [(pid, id) for zero-length segments])
    '''
    bad_segments = list()
    for row in raw_data.iter_row():
        idx = int(row[COLS.ID])
        pid = int(row[COLS.P])
        if pid != ROOT_ID:
            prow = raw_data.get_row(pid)
            if (is_neurite_segment((prow, row)) and
                    np.all(row[COLS.X: COLS.R] == prow[COLS.X: COLS.R])):
                bad_segments.append((pid, idx))
    return len(bad_segments) == 0, bad_segments



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Mathematical and geometrical functions used to compute morphometrics'''
import math
from itertools import combinations
import numpy as np
from itertools import islice, izip
from neurom.core.dataformat import COLS


def vector(p1, p2):
    '''compute vector between two 3D points

    Args:
        p1, p2: indexable objects with
        indices 0, 1, 2 corresponding to 3D cartesian coordinates.

    Returns:
        3-vector from p1 - p2
    '''
    return np.subtract(p1[0:3], p2[0:3])


def point_dist2(p1, p2):
    '''compute the square of the euclidian distance between two 3D points

    Args:
        p1, p2: indexable objects with
        indices 0, 1, 2 corresponding to 3D cartesian coordinates.
    Returns:
        The square of the euclidian distance between the points.
    '''
    v = vector(p1, p2)
    return np.dot(v, v)


def point_dist(p1, p2):
    '''compute the euclidian distance between two 3D points

    Args:
        p1, p2: indexable objects with
        indices 0, 1, 2 corresponding to 3D cartesian coordinates.
    Returns:
        The euclidian distance between the points.
    '''
    return np.linalg.norm(vector(p1, p2))


def angle_3points(p0, p1, p2):
    ''' compute the angle in radians between three 3D points

    Calculated as the angle between p1-p0 and p2-p0.

    Args:
        p0, p1, p2:  indexable objects with
        indices 0, 1, 2 corresponding to 3D cartesian coordinates.

    Returns:
        Angle in radians between (p1-p0) and (p2-p0).
        nan if p0==p1 or p0==p2.
    '''
    vec1 = vector(p1, p0)
    vec2 = vector(p2, p0)
    return math.acos(np.dot(vec1, vec2) /
                     (np.linalg.norm(vec1) * np.linalg.norm(vec2)))


def polygon_diameter(points):
    ''' Compute the maximun euclidian distance between any two points
    in a list of points
    '''
    return max(point_dist(p0, p1) for (p0, p1) in combinations(points, 2))


def average_points_dist(p0, p_list):
    """
    Computes the average distance between a list of points
    and a given point p0.
    """
    return np.mean(list(point_dist(p0, p1) for p1 in p_list))


def path_distance(points):
    """
    Compute the path distance from given set of points
    """
    return sum(point_dist(p[0], p[1]) for p in izip(points, islice(points, 1, None)))


def segment_length(seg):
    '''Return the length of a segment.

    Returns: Euclidian distance between centres of points in seg
    '''
    return point_dist(seg[0], seg[1])


def segment_radius(seg):
    '''Return the mean radius of a segment

    Returns: arithmetic mean of the radii of the points in seg
    '''
    return (seg[0][COLS.R] + seg[1][COLS.R]) / 2.


def segment_radial_dist(seg, pos):
    '''Return the radial distance of a tree segment to a given point

    The radial distance is the euclidian distance between the mid-point of
    the segment and the point in question.

    Parameters:
        seg: tree segment

        pos: origin to which disrances are measured. It must have at lease 3
        components. The first 3 components are (x, y, z).
    '''
    return point_dist(pos, np.divide(np.add(seg[0], seg[1]), 2.0))


def segment_area(seg):
    '''Compute the surface area of a segment.

    Approximated as a conical frustum. Does not include the surface area
    of the bounding circles.
    '''
    r0 = seg[0][COLS.R]
    r1 = seg[1][COLS.R]
    h2 = point_dist2(seg[0], seg[1])
    return math.pi * (r0 + r1) * math.sqrt((r0 - r1) ** 2 + h2)


def segment_volume(seg):
    '''Compute the volume of a segment.

    Approximated as a conical frustum.
    '''
    r0 = seg[0][COLS.R]
    r1 = seg[1][COLS.R]
    h = point_dist(seg[0], seg[1])
    return math.pi * h * ((r0 * r0) + (r0 * r1) + (r1 * r1)) / 3.0


def taper_rate(p0, p1):
    '''Compute the taper rate between points p0 and p1

    Args:
        p0, p1: iterables with first 4 components containing (x, y, z, r)

    Returns:
        The taper rate, defined as the absolute value of the difference in
        the diameters of p0 and p1 divided by the euclidian distance
        between them.
    '''
    return 2 * abs(p0[COLS.R] - p1[COLS.R]) / point_dist(p0, p1)


def segment_taper_rate(seg):
    '''Compute the taper rate of a segment

    Returns:
        The taper rate, defined as the absolute value of the difference in
        the diameters of the segment's two points divided by the euclidian
        distance between them.
    '''
    return taper_rate(seg[0], seg[1])


# Useful alias for path_distance
section_length = path_distance



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Generic tree class and iteration functions'''
from itertools import chain, imap, ifilter, repeat


class Tree(object):
    '''
    Simple tree class. This is a recursive data structure, with each tree
    holding a value and a list of children trees. Every node is a tree.
    '''
    def __init__(self, value):
        self.value = value
        self.parent = None
        self.children = list()

    def add_child(self, tree):
        '''Add a child to the list of this tree's children

        This tree becomes the added tree's parent
        '''
        self.children.append(tree)
        tree.parent = self
        return tree


def is_forking_point(tree):
    '''Is tree a forking point?'''
    return len(tree.children) > 1


def is_bifurcation_point(tree):
    '''Is tree a bifurcation point?'''
    return len(tree.children) == 2


def is_leaf(tree):
    '''Is tree a leaf?'''
    return len(tree.children) == 0


def is_root(tree):
    '''Is tree the root node?'''
    return tree.parent is None


def ipreorder(tree):
    '''Depth-first pre-order iteration of tree nodes'''
    yield tree
    for v in chain(*imap(ipreorder, tree.children)):
        yield v


def ipostorder(tree):
    '''Depth-first post-order iteration of tree nodes'''
    for v in chain(*imap(ipostorder, tree.children)):
        yield v
    yield tree


def iupstream(tree):
    '''Iterate from a tree node to the root nodes'''
    t = tree
    while t is not None:
        yield t
        t = t.parent


def ileaf(tree):
    '''Iterator to all leaves of a tree'''
    return ifilter(is_leaf, ipreorder(tree))


def iforking_point(tree, iter_mode=ipreorder):
    '''Iterator to forking points. Returns a tree object.

    Parameters:
        tree: the tree over which to iterate
        iter_mode: iteration mode. Default: ipreorder.
    '''
    return ifilter(is_forking_point, iter_mode(tree))


def ibifurcation_point(tree, iter_mode=ipreorder):
    '''Iterator to bifurcation points. Returns a tree object.

    Parameters:
        tree: the tree over which to iterate
        iter_mode: iteration mode. Default: ipreorder.
    '''
    return ifilter(is_bifurcation_point, iter_mode(tree))


def isegment(tree, iter_mode=ipreorder):
    '''Iterate over segments

    Segments are parent-child pairs, with the child being the
    center of the iteration

    Parameters:
        tree: the tree over which to iterate
        iter_mode: iteration mode. Default: ipreorder.
    '''
    return imap(lambda t: (t.parent, t),
                ifilter(lambda t: t.parent is not None, iter_mode(tree)))


def isection(tree, iter_mode=ipreorder):
    '''Iterator to sections of a tree.

    Resolves to a tuple of sub-trees forming a section.

    Parameters:
        tree: the tree over which to iterate
        iter_mode: iteration mode. Default: ipreorder.
    '''
    def get_section(tree):
        '''get the upstream section starting from this tree'''
        ui = iupstream(tree)
        sec = [ui.next()]
        for i in ui:
            sec.append(i)
            if is_forking_point(i) or is_root(i):
                break
        sec.reverse()
        return tuple(sec)

    def seed_node(n):
        '''Is this node a good seed for upstream section finding?'''
        return not is_root(n) and (is_leaf(n) or is_forking_point(n))

    return imap(get_section,
                ifilter(seed_node, iter_mode(tree)))


def itriplet(tree):
    '''Iterate over triplets

    Post-order iteration yielding tuples with three consecutive sub-trees
    '''
    return chain(
        *imap(lambda n: zip(repeat(n.parent), repeat(n), n.children),
              ifilter(lambda n: not is_root(n) and not is_leaf(n),
                      ipreorder(tree))))


def i_branch_end_points(fork_point):
    '''Iterate over the furthest points in forking sections

    Parameters:
        fork_point: A tree object which is a forking point

    Returns:
        An iterator with end-points of the sections forking out\
            of fork_point
    '''

    def next_end_point(tree):
        '''Get the next node of a tree which is a section end-point
        '''
        i = ipreorder(tree)
        while True:
            n = i.next()
            if is_section_end_point(n):
                break
        return n

    def is_section_end_point(tree):
        '''Is this tree a section end point???
        '''
        return (not is_root(tree)) and (is_forking_point(tree) or is_leaf(tree))

    return imap(next_end_point, fork_point.children)


def val_iter(tree_iterator):
    '''Iterator adaptor to iterate over Tree.value'''
    def _deep_map(f, data):
        '''Recursive map function. Maintains type of iterables'''
        return (type(data)(_deep_map(f, x) for x in data)
                if hasattr(data, '__iter__')
                else f(data))
    return imap(lambda t: _deep_map(lambda n: n.value, t), tree_iterator)


def imap_val(f, tree_iterator):
    '''Map function f to value of tree_iterator's target
    '''
    return imap(f, val_iter(tree_iterator))


def i_chain(trees, iterator_type, mapping=None, tree_filter=None):
    '''Returns a mapped iterator to a collection of trees

    Provides access to all the elements of all the trees
    in one iteration sequence.

    Parameters:
        trees: iterator or iterable of tree objects
        iterator_type: type of the iteration (segment, section, triplet...)
        mapping: optional function to apply to the iterator's target.
        tree_filter: optional top level filter on properties of tree objects.
    '''
    nrt = (trees if tree_filter is None
           else filter(tree_filter, trees))

    chain_it = chain(*imap(iterator_type, nrt))
    return chain_it if mapping is None else imap_val(mapping, chain_it)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#

# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

''' Neuron class with basic analysis and plotting capabilities. '''

from neurom.io.utils import load_neuron
from neurom.core.types import TreeType
from neurom.core.types import checkTreeType
from neurom.core.tree import ipreorder
from neurom.core.tree import isection
from neurom.core.tree import isegment
from neurom.core.tree import i_chain as i_neurites
from neurom.analysis.morphmath import section_length
from neurom.analysis.morphmath import segment_length
from neurom.analysis.morphtree import set_tree_type
from neurom.analysis.morphtree import i_local_bifurcation_angle
from neurom.analysis.morphtree import i_remote_bifurcation_angle
from neurom.analysis.morphtree import i_section_radial_dist
from neurom.analysis.morphtree import i_section_path_length
from neurom.analysis.morphtree import n_sections
import math
import numpy as np


class Neuron(object):
    '''Class with basic analysis and plotting functionality

    By default returns iterables as numpy.arrays and applies no filtering.

    Arguments:
        filename: path to morphology file to be loaded.
        iterable_type: type of iterable to return from methods returning \
            collections (e.g list, tuple, numpy.array).

    Raises:
        neurom.exceptions.SomaError if no soma can be built from the data
        neurom.exceptions.NonConsecutiveIDsError if data IDs not consecutive

    Example:
        get the segment lengths of all apical dendrites in a neuron morphology.

    >>> from neurom import ezy
    >>> nrn = ezy.Neuron('test_data/swc/Neuron.swc')
    >>> nrn.get_segment_lengths(ezy.TreeType.apical_dendrite)

    Example:
        iterate over the segment surface areas of all axons in a neuron\
            morphology.

    >>> from neurom import ezy
    >>> from neurom.analysis import morphmath as mm
    >>> nrn = ezy.Neuron('test_data/swc/Neuron.swc')
    >>> for a in nrn.iter_segments(mm.segment_area, ezy.TreeType.axon):
          print (a)

    Example:
        calculate the mean volume of all neurite segments on a neuron\
            morphology

    >>> from neurom import ezy
    >>> from neurom.analysis import morphmath as mm
    >>> import numpy as np
    >>> nrn = ezy.Neuron('test_data/swc/Neuron.swc')
    >>> mv = np.mean([v for v in nrn.iter_segments(mm.segment_volume)])

    Example:
        use lists instead of numpy arrays and get section \
    lengths for the axon. Read an HDF5 v1 file:

    >>> from neurom import ezy
    >>> nrn = ezy.Neuron('test_data/h5/v1/Neuron.h5', iterable_type=list)
    >>> nrn.get_section_lengths(ezy.TreeType.axon)

    Example:
        plot the neuron and save to a file

    >>> fig, ax = nrn.plot()
    >>> fig.show()
    >>> fig.savefig('nrn.png')

    '''

    def __init__(self, filename, iterable_type=np.array):
        self._iterable_type = iterable_type
        self._nrn = load_neuron(filename, set_tree_type)
        self.soma = self._nrn.soma
        self.neurites = self._nrn.neurites

    def get_section_lengths(self, neurite_type=TreeType.all):
        '''Get an iterable containing the lengths of all sections of a given type'''
        return self._pkg(self.iter_sections(section_length, neurite_type))

    def get_segment_lengths(self, neurite_type=TreeType.all):
        '''Get an iterable containing the lengths of all segments of a given type'''
        return self._pkg(self.iter_segments(segment_length, neurite_type))

    def get_soma_radius(self):
        '''Get the radius of the soma'''
        return self._nrn.soma.radius

    def get_soma_surface_area(self):
        '''Get the surface area of the soma.

        Note:
            The surface area is calculated by assuming the soma is spherical.
        '''
        return 4 * math.pi * self.get_soma_radius() ** 2

    def get_local_bifurcation_angles(self, neurite_type=TreeType.all):
        '''Get local bifircation angles of all segments of a given type

        The local bifurcation angle is defined as the angle between
        the first segments of the bifurcation branches.

        Returns:
            Iterable containing bifurcation angles in radians
        '''
        return self.neurite_loop(i_local_bifurcation_angle,
                                 neurite_type=neurite_type)

    def get_remote_bifurcation_angles(self, neurite_type=TreeType.all):
        '''Get remote bifircation angles of all segments of a given type

        The remote bifurcation angle is defined as the angle between
        the lines joining the bifurcation point to the last points
        of the bifurcated sections.

        Returns:
            Iterable containing bifurcation angles in radians
        '''
        return self.neurite_loop(i_remote_bifurcation_angle,
                                 neurite_type=neurite_type)

    def get_section_radial_distances(self, origin=None, use_start_point=False,
                                     neurite_type=TreeType.all):
        '''Get an iterable containing section radial distances to origin of\
           all neurites of a given type

        Parameters:
            origin: Point wrt which radial dirtance is calulated\
                    (default tree root)
            use_start_point: if true, use the section's first point,\
                             otherwise use the end-point (default False)
            neurite_type: Type of neurites to be considered (default all)
        '''
        return self.neurite_loop(lambda t: i_section_radial_dist(t, origin,
                                                                 use_start_point),
                                 neurite_type=neurite_type)

    def get_section_path_distances(self, use_start_point=False,
                                   neurite_type=TreeType.all):
        '''
        Get section path distances of all neurites of a given type
        The section path distance is measured to the neurite's root.

        Parameters:
            use_start_point: boolean\
            if true, use the section's first point,\
            otherwise use the end-point (default False)
            neurite_type: TreeType\
            Type of neurites to be considered (default all)

        Returns:
            Iterable containing the section path distances.
        '''
        return self.neurite_loop(lambda t: i_section_path_length(t, use_start_point),
                                 neurite_type=neurite_type)

    def get_n_sections(self, neurite_type=TreeType.all):
        '''Get the number of sections of a given type'''
        return sum(n_sections(t) for t in self._nrn.neurites
                   if checkTreeType(neurite_type, t.type))

    def get_n_sections_per_neurite(self, neurite_type=TreeType.all):
        '''Get an iterable with the number of sections for a given neurite type'''
        return self._iterable_type(
            [n_sections(n) for n in self._nrn.neurites
             if checkTreeType(neurite_type, n.type)]
        )

    def get_n_neurites(self, neurite_type=TreeType.all):
        '''Get the number of neurites of a given type in a neuron'''
        return sum(1 for n in self._nrn.neurites
                   if checkTreeType(neurite_type, n.type))

    def iter_neurites(self, iterator_type, mapping=None, neurite_type=TreeType.all):
        '''Iterate over collection of neurites applying iterator_type

        Parameters:
            iterator_type: Type of iterator with which to perform the iteration.\
            (e.g. isegment, isection, i_section_path_length)
            mapping: mapping function to be applied to the target of iteration.\
            (e.g. segment_length). Must be compatible with the iterator_type.
            neurite_type: TreeType object. Neurites of incompatible type are\
            filtered out.

        Returns:
            Iterator of mapped iteration targets.

        Example:
            Get the total volume of all neurites in the cell and the total\
                length or neurites from their segments.

        >>> from neurom import ezy
        >>> from neurom.analysis import morphmath as mm
        >>> from neurom.core import tree as tr
        >>> nrn = ezy.Neuron('test_data/swc/Neuron.swc')
        >>> v = sum(nrn.iter_neurites(tr.isegment, mm.segment_volume))
        >>> tl = sum(nrn.iter_neurites(tr.isegment, mm.segment_length)))

        '''
        return i_neurites(self._nrn.neurites,
                          iterator_type,
                          mapping,
                          tree_filter=lambda t: checkTreeType(neurite_type,
                                                              t.type))

    def iter_points(self, mapfun, neurite_type=TreeType.all):
        '''Iterator to neurite points with mapping

        Parameters:
            mapfun: mapping function to be applied to points.
            neurite_type: type of neurites to iterate over.
        '''
        return self.iter_neurites(ipreorder, mapfun, neurite_type)

    def iter_segments(self, mapfun, neurite_type=TreeType.all):
        '''Iterator to neurite segments with mapping

        Parameters:
            mapfun: mapping function to be applied to segments.
            neurite_type: type of neurites to iterate over.
        '''
        return self.iter_neurites(isegment, mapfun, neurite_type)

    def iter_sections(self, mapfun, neurite_type=TreeType.all):
        '''Iterator to neurite sections with mapping

        Parameters:
            mapfun: mapping function to be applied to sections.
            neurite_type: type of neurites to iterate over.
        '''
        return self.iter_neurites(isection, mapfun, neurite_type)

    def neurite_loop(self, iterator_type, mapping=None, neurite_type=TreeType.all):
        '''Iterate over collection of neurites applying iterator_type

        Parameters:
            iterator_type: Type of iterator with which to perform the iteration.
            (e.g. isegment, isection, i_section_path_length)
            mapping: mapping function to be applied to the target of iteration.
            (e.g. segment_length). Must be compatible with the iterator_type.
            neurite_type: TreeType object. Neurites of incompatible type are
            filtered out.

        Returns:
            Iterable containing the iteration targets after mapping.
        '''
        return self._pkg(self.iter_neurites(iterator_type, mapping, neurite_type))

    def _pkg(self, iterator):
        '''Create an iterable from an iterator'''
        return self._iterable_type([i for i in iterator])



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Neuron classes and functions'''
from neurom.analysis.morphmath import average_points_dist
import neurom.core.tree as tr
from neurom.core.dataformat import COLS
from neurom.exceptions import SomaError
import numpy as np


class SOMA_TYPE(object):
    '''Enumeration holding soma types

    * Type A: single point at centre
    * Type B: Three points on circumference of sphere
    * Type C: More than three points
    * INVALID: Not satisfying any of the above
    '''
    INVALID, A, B, C = xrange(4)

    @staticmethod
    def get_type(points):
        '''get the type of the soma'''
        npoints = len(points)
        return {0: SOMA_TYPE.INVALID,
                1: SOMA_TYPE.A,
                3: SOMA_TYPE.B,
                2: SOMA_TYPE.INVALID}.get(npoints, SOMA_TYPE.C)


class BaseSoma(object):
    '''Base class for a soma.

    Holds a list of raw data rows corresponding to soma points
    and provides iterator access to them.
    '''
    def __init__(self, points):
        self._points = points

    def iter(self):
        '''Iterator to soma contents'''
        return iter(self._points)


class SomaA(BaseSoma):
    '''
    Type A: 1point soma
    Represented by a single point.
    '''
    def __init__(self, points):
        super(SomaA, self).__init__(points)
        self.center = tuple(points[0][:COLS.R])
        self.radius = points[0][COLS.R]


class SomaB(BaseSoma):
    '''
    Type B: 3point soma
    Represented by 3 points.
    Reference: neuromorpho.org
    The first point constitutes the center of the soma.
    An equivalent radius (rs) is computed as the average distance
    of the other two points.
    '''
    def __init__(self, points):
        super(SomaB, self).__init__(points)
        self.center = tuple(points[0][:COLS.R])
        self.radius = average_points_dist(points[0], (points[1], points[2]))


class SomaC(BaseSoma):
    '''
    Type C: multiple points soma
    Represented by a contour.
    Reference: neuromorpho.org
    The first point constitutes the center of the soma,
    with coordinates (xs, ys, zs) corresponding to the
    average of all the points in the single contour.
    An equivalent radius (rs) is computed as the average distance
    of each point of the single contour from this center.
    '''
    def __init__(self, points):
        super(SomaC, self).__init__(points)
        self.center = tuple(points[0][:COLS.R])
        self.radius = average_points_dist(points[0], points[1:])


def make_soma(points):
    '''Make a soma object from a set of points

    Infers the soma type (SomaA, SomaB or SomaC) from the points.

    Raises:
        SomaError if no soma points found or points incompatible with soma.
    '''
    stype = SOMA_TYPE.get_type(points)
    if stype == SOMA_TYPE.INVALID:
        raise SomaError('Invalid soma points')

    return {SOMA_TYPE.A: SomaA,
            SOMA_TYPE.B: SomaB,
            SOMA_TYPE.C: SomaC}[stype](points)


class Neuron(object):
    '''Toy neuron class for testing ideas'''
    def __init__(self, soma_points, neurites, name='Neuron'):
        '''Construct a Neuron

        Arguments:
            soma_points: iterable of soma points.
            neurites: iterable of neurite tree structures.
            name: Optional name for this Neuron.

        Raises:
            SomaError if soma can't be built from soma_points.
        '''
        self.soma = make_soma(soma_points)
        self.neurites = neurites
        self.id = name

    def i_neurites(self, iterator_type, mapping=None, tree_filter=None):
        '''Returns a mapped iterator to all the neuron's neurites

        Provides access to all the elements of all the neurites
        in one iteration sequence.

        Parameters:
            iterator_type: type of the iteration (segment, section, triplet...)
            mapping: optional function to apply to the iterator's target.
            tree_filter: optional top level filter on properties of neurite tree objects.
        '''
        return tr.i_chain(self.neurites, iterator_type, mapping, tree_filter)


def bounding_box(nrn):
    '''Return 3D bounding box of a neuron

    Returns:
        2D numpy array of [[min_x, min_y, min_z], [max_x, max_y, max_z]]
    '''

    # Get the bounding coordinates of the neurites
    nmin_xyz, nmax_xyz = (np.array([np.inf, np.inf, np.inf]),
                          np.array([np.NINF, np.NINF, np.NINF]))

    for p in tr.i_chain(nrn.neurites, tr.ipreorder, lambda p: p):
        nmin_xyz = np.minimum(p[:COLS.R], nmin_xyz)
        nmax_xyz = np.maximum(p[:COLS.R], nmax_xyz)

    # Get the bounding coordinates of the soma
    smin_xyz = np.array(nrn.soma.center) - nrn.soma.radius

    smax_xyz = np.array(nrn.soma.center) + nrn.soma.radius

    return np.array([np.minimum(smin_xyz, nmin_xyz),
                     np.maximum(smax_xyz, nmax_xyz)])



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

""" Distribution configuration for neurom
"""
# pylint: disable=R0801
import os
from setuptools import setup
from setuptools import find_packages
import pip
from pip.req import parse_requirements
from optparse import Option
from neurom.version import VERSION


def parse_reqs(reqs_file):
    ''' parse the requirements '''
    options = Option('--workaround')
    options.skip_requirements_regex = None
    # Hack for old pip versions
    # Versions greater than 1.x have a required parameter "sessions" in
    # parse_requierements
    if pip.__version__.startswith('1.'):
        install_reqs = parse_requirements(reqs_file, options=options)
    else:
        from pip.download import PipSession  # pylint:disable=E0611
        options.isolated_mode = False
        install_reqs = parse_requirements(reqs_file,  # pylint:disable=E1123
                                          options=options,
                                          session=PipSession)

    return [str(ir.req) for ir in install_reqs]


BASEDIR = os.path.dirname(os.path.abspath(__file__))
REQS = parse_reqs(os.path.join(BASEDIR, 'requirements.txt'))

EXTRA_REQS_PREFIX = 'requirements_'
EXTRA_REQS = {}

for file_name in os.listdir(BASEDIR):
    if not file_name.startswith(EXTRA_REQS_PREFIX):
        continue
    base_name = os.path.basename(file_name)
    (extra, _) = os.path.splitext(base_name)
    extra = extra[len(EXTRA_REQS_PREFIX):]
    EXTRA_REQS[extra] = parse_reqs(file_name)

config = {
    'description': 'NeuroM: a light-weight neuron morphology analysis package',
    'author': 'BBP Algorithm Development Team',
    'url': 'http://www.example.com',
    'author_email': 'juan.palacios@epfl.ch, lida.kanari@epfl.ch',
    'version': VERSION,
    'install_requires': REQS,
    'extras_require': EXTRA_REQS,
    'packages': find_packages(),
    'scripts': ['apps/raw_data_check', 'apps/morph_check'],
    'name': 'neurom',
    'include_package_data': True,
}

setup(**config)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

''' Module for morphology data loading and access

Data is unpacked into a 2-dimensional raw data block:

    [X, Y, Z, R, TYPE, ID, PARENT_ID]

There is one such row per measured point.

Functions to umpack the data and a higher level wrapper are provided. See

* load_data
* load_swc
* load_h5v1
* load_h5v2
* RawDataWrapper
'''
import os
from collections import defaultdict
from itertools import ifilter
from neurom.core.point import as_point
from neurom.core.dataformat import COLS
from neurom.core.dataformat import ROOT_ID
from neurom.io.swc import SWC
from neurom.io.hdf5 import H5


def load_data(filename):
    '''Unpack filename and return a RawDataWrapper object containing the data

    Determines format from extension. Currently supported:

        * SWC (case-insensitive extension ".swc")
        * H5 v1 and v2 (case-insensitive extension ".h5"). Attempts to
          determine the version from the contents of the file.
    '''
    def unpack_data():
        '''Read an SWC file and return a tuple of data, offset, format.
        Forwards filename to appropriate reader depending on extension'''
        _READERS = {'swc': SWC.read, 'h5': H5.read}
        extension = os.path.splitext(filename)[1][1:]
        return _READERS[extension.lower()](filename)

    return RawDataWrapper(unpack_data())


def load_h5v1(filename):
    '''Unpack HDF5 v1 file  and return a RawDataWrapper object
    '''
    return RawDataWrapper(H5.read_v1(filename))


def load_h5v2(filename, stage='raw'):
    '''Unpack HDF5 v2 file  and return a RawDataWrapper object
    '''
    return RawDataWrapper(H5.read_v2(filename, stage))


def load_swc(filename):
    '''Unpack SWC file  and return a RawDataWrapper object
    '''
    return RawDataWrapper(SWC.read(filename))


class RawDataWrapper(object):
    '''Class holding an array of data and an offset to the first element
    and giving basic access to its elements

    The array contains rows with
        [X, Y, Z, R, TYPE, ID, PID]

    where the elements are

    * X, Y, Z: X, Y, Z coordinates of point
    * R: Radius of node at that point
    * TYPE: Type of neuronal segment.
    * ID: Identifier for a point. Non-negative, increases by one for each row.
    * PID: ID of parent point
    '''
    def __init__(self, raw_data):
        self.data_block, self._offset, self.fmt = raw_data
        self.adj_list = defaultdict(list)
        for row in self.data_block:
            self.adj_list[int(row[COLS.P])].append(int(row[COLS.ID]))

    def get_children(self, idx):
        ''' get list of ids of children of parent with id idx'''
        if idx != ROOT_ID and idx not in self.get_ids():
            raise LookupError('Invalid id: {0}'.format(idx))
        return self.adj_list[idx]

    def _apply_offset(self, idx):
        ''' Apply global offset to an id'''
        return idx - self._offset

    def get_parent(self, idx):
        '''get the parent of element with id idx'''
        if idx not in self.get_ids():
            raise LookupError('Invalid id: {0}'.format(idx))
        return int(self.data_block[self._apply_offset(idx)][COLS.P])

    def get_point(self, idx):
        '''Get point data for element idx'''
        idx = self._apply_offset(idx)
        p = as_point(self.data_block[idx]) if idx > ROOT_ID else None
        return p

    def get_row(self, idx):
        '''Get row from idx'''
        idx = self._apply_offset(idx)
        return self.data_block[idx] if idx > ROOT_ID else None

    def get_col(self, col_id):
        '''Get column from ID'''
        return self.data_block[:, col_id]

    def get_end_points(self):
        ''' get the end points of the tree

        End points have no children so are not in the
        adjacency list.
        '''
        return [int(i) for i in
                set(self.get_col(COLS.ID)) - set(self.adj_list.keys())]

    def get_ids(self, pred=None):
        '''Get the list of ids for rows satisfying an optional row predicate'''
        return [r[COLS.ID] for r in self.iter_row(None, pred)]

    def get_fork_points(self):
        '''Get list of point ids for points with more than one child'''
        return [i for i, l in self.adj_list.iteritems() if len(l) > 1]

    def iter_row(self, start_id=None, pred=None):
        '''Get an row iterator to a starting at start_id and satisfying a
        row predicate pred.
        '''
        if start_id is None:
            start_id = self._offset

        start_id = self._apply_offset(start_id)
        if start_id < 0 or start_id >= self.data_block.shape[0]:
            raise LookupError('Invalid id: {0}'.format(start_id))

        irow = iter(self.data_block[start_id:])
        return irow if pred is None else ifilter(pred, irow)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Basic functions used for tree analysis

These functions all depend on the internal structure of the tree or its
different iteration modes.
'''
from itertools import imap
from neurom.core import tree as tr
from neurom.core.types import TreeType
import neurom.analysis.morphmath as mm
from neurom.core.dataformat import COLS
from neurom.core.tree import val_iter
import numpy as np
import logging

LOG = logging.getLogger(__name__)


def path_length(tree):
    '''Get the path length from a sub-tree to the root node'''
    return np.sum(s for s in
                  tr.imap_val(mm.segment_length, tr.isegment(tree, tr.iupstream)))


def local_bifurcation_angle(bifurcation_point):
    '''Return the opening angle between two out-going segments
    in a bifurcation point
    '''
    return mm.angle_3points(bifurcation_point.value,
                            bifurcation_point.children[0].value,
                            bifurcation_point.children[1].value)


def branch_order(tree_section):
    '''Branching order of a tree section

    The branching order is defined as the depth of the tree section.

    Note:
        The first level has branch order 0.
    '''
    node = tree_section[-1]
    bo = sum(1 for _ in tr.iforking_point(node, tr.iupstream))
    return bo - 2 if tr.is_forking_point(node) else bo - 1


def i_segment_length(tree):
    ''' return an iterator of tree segment lengths
    '''
    return tr.imap_val(mm.segment_length, tr.isegment(tree))


def i_segment_radius(tree):
    ''' return an iterator of tree segment radii
    '''
    return tr.imap_val(mm.segment_radius, tr.isegment(tree))


def i_segment_volume(tree):
    ''' return an iterator of tree segment volumes
    '''
    return tr.imap_val(mm.segment_volume, tr.isegment(tree))


def i_segment_area(tree):
    ''' return an iterator of tree segment areas
    '''
    return tr.imap_val(mm.segment_area, tr.isegment(tree))


def i_segment_radial_dist(pos, tree):
    '''Return an iterator of radial distances of tree segments to a given point

    The radial distance is the euclidian distance between the mid-point of
    the segment and the point in question.

    Parameters:
        pos: origin to which disrances are measured. It must have at least 3
        components. The first 3 components are (x, y, z).

        tree: tree of raw data rows.

    '''
    return tr.imap_val(lambda s: mm.segment_radial_dist(s, pos), tr.isegment(tree))


def i_segment_meander_angle(tree):
    '''Return an iterator to a tree meander angle

    The meander angle is defined as the angle between to adjacent  segments.
    Applies neurom.morphmath.angle_3points to triplets of
    '''

    return tr.imap_val(lambda t: mm.angle_3points(t[1], t[0], t[2]), tr.itriplet(tree))


def i_local_bifurcation_angle(tree):
    '''Return the opening angle between two out-going segments
    in a bifurcation point
    '''
    return imap(local_bifurcation_angle,
                tr.ibifurcation_point(tree))


def i_remote_bifurcation_angle(tree):
    '''Return the opening angle between the last segments of two out-going
    sections of a bifurcation point
    '''
    def _remangle(t):
        '''Helper to calculate the remote angle'''
        end_points = tuple(p for p in tr.i_branch_end_points(t))
        return mm.angle_3points(t.value, end_points[0].value, end_points[1].value)

    return imap(_remangle, tr.ibifurcation_point(tree))


def i_section_radial_dist(tree, pos=None, use_start_point=False):
    '''Return an iterator of radial distances of tree sections to a given point

    The radial distance is the euclidian distance between the either the
    end-point or rhe start point of the section and the point in question.

    Parameters:
        tree: tree object
        pos: origin to which distances are measured. It must have at least 3\
            components. The first 3 components are (x, y, z).\
            (default tree origin)

        use_start_point: If true, calculate distance from section start point,\
            else from end-point (default, False)

    '''
    pos = tree.value if pos is None else pos
    sec_idx = 0 if use_start_point else -1
    return tr.imap_val(lambda s: mm.point_dist(s[sec_idx], pos), tr.isection(tree))


def i_section_path_length(tree, use_start_point=False):
    '''Return an iterator of path lengths of tree sections

    Path lengths are measured to the tree's root.

    Parameters:
        tree: tree object
        use_start_point: If true, calculate path length from section start point,\
            else from end-point (default, False)

    '''
    sec_idx = 0 if use_start_point else -1
    return imap(lambda s: path_length(s[sec_idx]), tr.isection(tree))


def find_tree_type(tree):

    """
    Calculates the 'mean' type of the tree.
    Accepted tree types are:
    'undefined', 'axon', 'basal', 'apical'
    The 'mean' tree type is defined as the type
    that is shared between at least 51% of the tree's points.
    Returns:
        The type of the tree
    """

    tree_types = tuple(TreeType)

    types = [node[COLS.TYPE] for node in tr.val_iter(tr.ipreorder(tree))]
    types = [node[COLS.TYPE] for node in tr.val_iter(tr.ipreorder(tree))]

    return tree_types[int(np.median(types))]


def set_tree_type(tree):
    ''' Set the type of a tree as 'type' attribute'''
    tree.type = find_tree_type(tree)


def get_tree_type(tree):
    '''Return a tree's type

    If tree does not have a type, calculate it.
    Otherwise, use its pre-computed value.
    '''

    if not hasattr(tree, 'type'):
        set_tree_type(tree)

    return tree.type


def i_section_length(tree):
    """
    Return an iterator of tree section lengths
    """
    return tr.imap_val(mm.section_length, tr.isection(tree))


def n_sections(tree):
    """
    Return number of sections in tree
    """
    return sum(1 for _ in tr.isection(tree))


def get_bounding_box(tree):
    """
    Returns:
        The boundaries of the tree in three dimensions:
            [[xmin, ymin, zmin],
            [xmax, ymax, zmax]]
    """

    min_xyz, max_xyz = (np.array([np.inf, np.inf, np.inf]),
                        np.array([np.NINF, np.NINF, np.NINF]))

    for p in val_iter(tr.ipreorder(tree)):
        min_xyz = np.minimum(p[:COLS.R], min_xyz)
        max_xyz = np.maximum(p[:COLS.R], max_xyz)

    return np.array([min_xyz, max_xyz])
